#include <zephyr/logging/log.h>
#include <zephyr/zbus/zbus.h>

#include <stdio.h>
#include <string.h>

#include "driver-hook.h"
#include "utilities.h"
#include "ble-scanner.h"
#include "led-driver.h"
#include "servo-driver.h"
#include "ir-driver.h"
#include "wifi-controller.h"

#ifndef CONFIG_SOC_POSIX
LOG_MODULE_REGISTER(Driver_Hook, LOG_LEVEL_INF);
#else
LOG_MODULE_REGISTER(Driver_Hook, LOG_LEVEL_DBG);
#endif

ZBUS_CHAN_DECLARE(driverRequestChannel, driverResponseChannel, bleScanChannel);

static DriverHook *context;

ZBUS_LISTENER_DEFINE(driverHookListener, context->DriverRequestListener);

DriverHook::DriverHook(){
    context = this;
#ifndef CONFIG_SOC_POSIX
    WifiController *wifiController = new WifiController();
    bleScanner = new BleScanner();
    ledDriver = new LedDriver();
    servoDriver = new ServoDriver();
    irDriver = new IrDriver();
    irDriver->RegisterInterrupt(IrSensorCallback);
    bleScanner->RegisterCallback(BleDataCallback);
#endif

    zbus_chan_add_obs(&driverRequestChannel, &driverHookListener, K_NO_WAIT);
}

DriverHook::~DriverHook(){

}

void DriverHook::IrSensorCallback(uint8_t data){
    LOG_INF("Iterrupt triggeredd. pin %d", data);
    std::vector<uint8_t> sensorData;
    sensorData.push_back(data);

    DriverMessage_t sentMessage = {
        .number = IR_DRIVER,
        .type = REQUEST_SET,
        .value = sensorData
    };

    zbus_chan_pub(&driverResponseChannel, &sentMessage, K_FOREVER);
}

void DriverHook::BleDataCallback(std::string addr, std::vector<uint8_t> data, int8_t rssi){
    // LOG_HEXDUMP_INF(data.data(), data.size(), "BLE data:");
    LOG_DBG("ble driver hook");
    BleData_t sentMessage = {
		.address = addr,
		.data = data
	};
	zbus_chan_pub(&bleScanChannel, &sentMessage, K_MSEC(200));
}

void DriverHook::Init(){

}

int DriverHook::Deinit(){
    return 0;
}

void DriverHook::DriverRequestListener(const struct zbus_channel *chan){
    const DriverMessage_t *msg;
    if (&driverRequestChannel == chan) {
			msg = (DriverMessage_t *)zbus_chan_const_msg(chan); // Direct message access
			LOG_DBG("From listener -> driver =%d, value 0=%d", msg->number, msg->value[0]);
	}
    else{
        return;
    }
    
    std::vector<uint8_t> data;
    data.push_back(msg->number);

    DriverMessage_t sentMessage = {
        .number = msg->number,
        .type = msg->type,
    };
#ifndef CONFIG_SOC_POSIX
    switch (msg->number)
    {
    case BLUETOOTH:
        if(msg->type == REQUEST_SET){
            context->bleScanner->SetScan(msg->value[0]);
            sentMessage.value.push_back(0);
        }
        break;

    case LED:
        if(msg->type == REQUEST_SET){
            context->ledDriver->SetState(msg->value[0]);
        }
        else if(msg->type == REQUEST_GET){
            bool ledState = context->ledDriver->GetState();
            sentMessage.value.push_back(ledState);
        }
        break;

    case SERVO:
        if(msg->type == REQUEST_SET){
            context->servoDriver->SetPwm(msg->value[0]);
        }
        else if(msg->type == REQUEST_GET){
            uint8_t servoPwm = context->servoDriver->GetPwm();
            sentMessage.value.push_back(servoPwm);
        }
        break;
    
    default:
        LOG_ERR("Invalid Sensor %d", msg->number);
        break;
    }
#else
    static bool ledState;
    static uint8_t pwmState;
    switch (msg->number)
    {
    case BLUETOOTH:
        if(msg->type == REQUEST_SET){
            sentMessage.value.push_back(0);
        }
        break;

    case LED:
        if(msg->type == REQUEST_SET){
            ledState = msg->value[0];
            sentMessage.value.push_back(ledState);
        }
        else if(msg->type == REQUEST_GET){
            sentMessage.value.push_back(ledState);
        }
        break;

    case SERVO:
        if(msg->type == REQUEST_SET){
            pwmState = msg->value[0];
        }
        else if(msg->type == REQUEST_GET){
            sentMessage.value.push_back(pwmState);
        }
        break;
    
    default:
        LOG_ERR("Invalid Sensor %d", msg->number);
        break;
    }
#endif

    int status = zbus_chan_pub(&driverResponseChannel, &sentMessage, K_MSEC(200));
    LOG_DBG("Status pub %d", status);
}