#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include <zephyr/zbus/zbus.h>

#ifndef CONFIG_SOC_POSIX
#include <zephyr/net/socket.h>
#endif

#include <stdio.h>
#include <string.h>

#include "debug-handler.h"
#include "utilities.h"

#define DEBUG_THREAD_STACK_SIZE 2048
#define DEBUG_THREAD_PRIORITY 5

#define PORT 12345

#define PASS 0x41, 0x42, 0x43

LOG_MODULE_REGISTER(Debug_Handler);
ZBUS_CHAN_DECLARE(driverResponseChannel, stateChannel, bleScanChannel);

K_THREAD_STACK_DEFINE(debugThreadStack, DEBUG_THREAD_STACK_SIZE);

// Thread and kernels
struct k_thread debugThreadHandler;
k_tid_t debugThreadId;
static DebugHandler *context;

// TCP
int serv;
int client;
bool isDebugMode = false;

// Debug Client
bool isDriverAvailable = false;
bool isBleAvailable = false;
DebugMessage_t debugData; 

ZBUS_LISTENER_DEFINE(debugDriverListener, context->DriverResponseListener);
ZBUS_LISTENER_DEFINE(stateListener, context->StateListener);
ZBUS_LISTENER_DEFINE(bleListener, context->BleListener);

DebugHandler::DebugHandler(){
	context = this;
    debugThreadId = k_thread_create(&debugThreadHandler, debugThreadStack, DEBUG_THREAD_STACK_SIZE, 
									  (k_thread_entry_t) DebugThread, 
									  this, NULL, NULL, 
									  DEBUG_THREAD_PRIORITY, 0, K_NO_WAIT);
	zbus_chan_add_obs(&driverResponseChannel, &debugDriverListener, K_NO_WAIT);
	zbus_chan_add_obs(&stateChannel, &stateListener, K_NO_WAIT);
	zbus_chan_add_obs(&bleScanChannel, &bleListener, K_NO_WAIT);
}

void DebugHandler::Init(){

}

int DebugHandler::Deinit(){
    return 0;
}

void DebugHandler::DebugThread(DebugHandler *context){
#ifndef CONFIG_SOC_POSIX
    struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(PORT);
	addr.sin_addr = INADDR_ANY_INIT;

	serv = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (serv < 0) {
		LOG_ERR("error: socket: %d\n", errno);
		return;
	}

	if (bind(serv, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		LOG_ERR("error: bind: %d\n", errno);
		return;
	}

	if (listen(serv, 5) < 0) {
		LOG_ERR("error: listen: %d\n", errno);
		return;
	}

	while (1) {
		struct sockaddr_in client_addr;
		socklen_t client_addr_len = sizeof(client_addr);
		char addr_str[32];
		client = accept(serv, (struct sockaddr *)&client_addr,
				    &client_addr_len);

		if (client < 0) {
			LOG_ERR("error: accept: %d\n", errno);
			continue;
		}

		inet_ntop(client_addr.sin_family, &client_addr.sin_addr,
			  addr_str, sizeof(addr_str));
		LOG_INF("Connection from %s\n", addr_str);

		while (1) {
			char buf[128], *p;
			int len = recv(client, buf, sizeof(buf), 0);
			int out_len;

			if (len <= 0) {
				if (len < 0) {
					LOG_ERR("error: recv: %d\n", errno);
				}
				break;
			}

			char pass[] = {PASS};
			if(memcmp(pass, buf, sizeof(pass)) == 0){
				LOG_INF("DEBUG ACTIVE");
				isDebugMode = true;
#if 0 //For testing purpose. will block indefinetely
				while(1){
					uint8_t dummy[2048];
					context->SendPacket(dummy, sizeof(dummy));
				}
#endif				
			}

			LOG_HEXDUMP_INF(buf, (uint32_t)len, "DATA TCP: ");
		}

		close(client);
		isDebugMode = false;
		LOG_INF("Connection to %s close", addr_str);
	}
#else
	while(1){
		std::vector<uint8_t> driverData;
		std::vector<uint8_t> bleData;
		uint8_t bleArray[128];
		uint8_t driverArray[128];
		uint16_t bleLen = 0;
		uint16_t driverLen = 0;
		
		if(isDriverAvailable){
			driverData.insert(driverData.end(), &driverArray[2], &driverArray[2] + driverLen); 
			DriverMessage_t sentMessage = {
				.number = (DriverNumb_e)driverArray[0],
				.type = (ExchangeType_e)driverArray[1],
				.value = driverData
			};
			zbus_chan_pub(&driverResponseChannel, &sentMessage, K_MSEC(200));
			isDriverAvailable = false;
		}

		if(isBleAvailable){
			bleData.insert(bleData.end(), &bleArray[0], &bleArray[bleLen]);
			BleData_t sentMessage = {
				.address = std::string(""), //temporary
				.data = bleData
			};
			zbus_chan_pub(&bleScanChannel, &sentMessage, K_MSEC(200));
			isBleAvailable = false;
		}
		k_msleep(100);
	}
#endif
}

void DebugHandler::DriverResponseListener(const struct zbus_channel *chan){
	const DriverMessage_t *driver;
	if (&driverResponseChannel == chan) {
			driver = (DriverMessage_t *)zbus_chan_const_msg(chan); // Direct message access
			// LOG_INF("From listener -> driver =%d, value 0=%d", driver->number, driver->value[0]);
			if(isDebugMode){
				std::vector<uint8_t> dataSent;
				uint16_t payloadSize = driver->value.size() + 2;
				dataSent.push_back(DEBUG_DRIVER);
				dataSent.push_back((payloadSize >> 8) & 0xff);
				dataSent.push_back(payloadSize & 0xff);
				dataSent.push_back(driver->number);
				dataSent.push_back(driver->type);
				copy(driver->value.begin(), driver->value.end(), back_inserter(dataSent));
				context->SendPacket((uint8_t *)dataSent.data(), dataSent.size());
			}
	}
}

void DebugHandler::StateListener(const struct zbus_channel *chan){
	const StateMessage_t *state;
	if (&stateChannel == chan) {
			state = (StateMessage_t *)zbus_chan_const_msg(chan); // Direct message access
			// LOG_INF("From listener -> state =%d, value 0=%d", state->flag, state->value[0]);
			if(isDebugMode){
				std::vector<uint8_t> dataSent;
				dataSent.push_back(DEBUG_STATE);
				dataSent.push_back((state->value.size() >> 8) & 0xff);
				dataSent.push_back(state->value.size() & 0xff);
				copy(state->value.begin(), state->value.end(), back_inserter(dataSent));
				context->SendPacket((uint8_t *)dataSent.data(), dataSent.size());
			}
	}
}

void DebugHandler::BleListener(const struct zbus_channel *chan){
	const BleData_t *ble;
	if (&bleScanChannel == chan) {
			ble = (BleData_t *)zbus_chan_const_msg(chan); // Direct message access
			// LOG_INF("Ble -> address =%s", ble->address.c_str());
			// LOG_HEXDUMP_INF(ble->data.data(), ble->data.size(), "data raw");
			if(isDebugMode){
				std::vector<uint8_t> dataSent;
				dataSent.push_back(DEBUG_BLE);
				dataSent.push_back((ble->data.size() >> 8) & 0xff);
				dataSent.push_back(ble->data.size() & 0xff);
				copy(ble->data.begin(), ble->data.end(), back_inserter(dataSent));
				context->SendPacket((uint8_t *)dataSent.data(), dataSent.size());
			}
	}
}

int DebugHandler::SendPacket(uint8_t *buf, uint16_t size){
#ifndef CONFIG_SOC_POSIX
	int sentlength, initialLength;
	initialLength = size;
	do {
		sentlength = send(client, buf, initialLength, 0);
		if (sentlength < 0) {
			LOG_ERR("error: send: %d\n", errno);
			break;
		}
		buf += sentlength;
		initialLength-= sentlength;
	} while (initialLength);
	return initialLength;
#else
	return 0;
#endif
}